#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import os
import numpy as np
import random
import time

from parlai.core.message import Message
from parlai.core.worlds import validate
import parlai.utils.logging as logging
from parlai.crowdsourcing.tasks.model_chat.utils import Compatibility
from parlai.crowdsourcing.tasks.model_chat.worlds import (
    get_bot_worker,
    ModelChatWorld,
    ModelChatOnboardWorld,
)

PERSONA_SETTER_AGENT = 'persona-agent'


class MultipartyChatWorld(ModelChatWorld):
    def __init__(self, opt, agent, bot, context_info=None):
        super(ModelChatWorld, self).__init__(opt, agent=agent, bot=bot)
        self.context_info = context_info
        self.personas = self.context_info['personas']

    def _run_initial_turn(self) -> None:
        """
        Run the initial turn for both the human and the bot.

        Optionally show the bot its persona. If we are in BST conversation mode, show 2
        previous BST utterances to both the human and the bot; if we are in Meena-like
        conversation mode, show "Hi!" to the human and the bot and let the bot respond
        accordingly.
        """
        # Removing the agents ids for use the id names for the agents.
        self.bot.agent_id = None
        self.agent.agent_id = None

        if self.opt['include_persona']:
            # Sending persona to the bot agent
            message = {"episode_done": False}
            assert isinstance(
                self.personas[1], dict
            ), 'Unknown persona format. Check the ContextGenerator in your task.'
            message['id'] = PERSONA_SETTER_AGENT
            message['text'] = 'PERSONA SETTING MESSAGE'
            message['personas'] = self.personas
            if 'location' in self.context_info:
                message['location'] = self.context_info['location']
            # The bot seeing its persona does not count as a "turn"
            self.bot.observe(validate(message), increment_turn=False)

            # Sending persona to the agent
            self.agent.observe(
                validate(
                    {
                        'episode_done': False,
                        'id': PERSONA_SETTER_AGENT,
                        'task_data': {
                            'personas': self.personas,
                            'location': self.context_info['location'],
                        },
                    }
                )
            )

    def has_final_rating(self, act):
        return act.get('task_data', {}).get('final_rating') is not None

    def parley(self):
        act = None  # Adding this for linter errors.
        logging.verbose(
            f'{self.__class__.__name__}:{self.tag}: is at turn {self.task_turn_idx}, with {self.num_turns} pairs of turns needed...'
        )

        if self.task_turn_idx == 0:
            self._run_initial_turn()
            self.task_turn_idx += 1
            return

        """Otherwise, we proceed accordingly"""
        logging.verbose(
            f'{self.__class__.__name__}:{self.tag}: About to act with task turn idx: {self.task_turn_idx}'
        )

        if not self.chat_done:
            act = self.bot.act()
            human_turn = act.get('human_turn', False)
            # Bot decided it is human turn.

            if human_turn:
                act = self.agent.act(timeout=self.max_resp_time)
                self.chat_done = self.has_final_rating(act)
                Compatibility.backward_compatible_force_set(
                    act, 'id', self.personas[0]['name']
                )

                act = Message(Compatibility.maybe_fix_act(act)).json_safe_payload()
            utterance_data = {
                'agent_idx': 0 if human_turn else 1,
                # Get rid of annotations HTML if it's the bot response
                'text': act['text'].split('<br>')[0],
                'id': act.get('id', 'NULL_ID'),
            }
            self.dialog.append(utterance_data)

            if human_turn:
                self.bot.observe(validate(act))
            else:
                act['needs_rating'] = True
                self.agent.observe(validate(act))

                # The new act replaces the old one
                act = self.agent.act(timeout=self.max_resp_time)
                act.force_set('text', 'THIS IS A RATING ACTION')
                p = act['task_data'].get('problem_data_for_prior_message')
                if p is not None:
                    turn_idx = -1
                    # Attach the problem data to the last utterance (just generated by bot).
                    self.__add_problem_data_to_utterance(p, turn_idx=turn_idx)

                self.chat_done = self.has_final_rating(act)

            self.task_turn_idx += 1

        if self.chat_done:
            self.dialog[-1]['final_rating'] = act['task_data']['final_rating']

            # Save the final chat data
            date_folder = time.strftime('%Y_%m_%d')
            time_string = time.strftime('%Y%m%d_%H%M%S')
            chat_data_subfolder = os.path.join(
                self.opt['chat_data_folder'], date_folder
            )
            os.makedirs(chat_data_subfolder, exist_ok=True)
            chat_data_path = os.path.join(
                chat_data_subfolder,
                f'{time_string}_{np.random.randint(0, 1000)}_{self.task_type}.json',
            )
            self.final_chat_data = self.get_final_chat_data()
            self.agent.mephisto_agent.state.messages.append(
                {
                    'final_chat_data': self.final_chat_data,
                    'data': {},
                    'packet_type': None,
                    'timestamp': None,
                }
            )
            # Append the chat data directly to the agent state's message list in
            # order to prevent the worker from seeing a new text response in the UI.
            # Add some dummy keys for compatibility with all agent state messages
            # TODO: remove this when no longer saving data to disk manually
            with open(chat_data_path, 'w+') as f_json:
                data_str = json.dumps(self.final_chat_data)
                f_json.write(data_str)
            logging.info(
                f'{self.__class__.__name__}:{self.tag}: Data saved at '
                f'{chat_data_path} for model: {self.bot.worker_id}.'
            )

            # Soft-block the worker if there were acceptability violations
            acceptability_violations = self.final_chat_data['acceptability_violations'][
                0
            ]
            if acceptability_violations is not None and acceptability_violations != '':
                logging.warning(
                    f'**NOTE** Acceptability violations detected: {acceptability_violations}'
                )
                # Grant the failed qualification
                self.agent.mephisto_agent.get_worker().grant_qualification(
                    self.block_qualification, 1
                )

    def __add_problem_data_to_utterance(self, p, turn_idx: int):
        """
        Attach problem data to the bot's prior utterance, given by turn_idx.

        This is copied exactly from the main model_chat world.
        """
        logging.verbose(f'Problem matrix:\n{p}')
        assert (
            self.dialog[turn_idx]['agent_idx'] == 1
        ), 'Problem data must be attached to a bot utterance.'
        assert (
            'problem_data' not in self.dialog[turn_idx]
        ), "Don't overwrite existing problem data!"
        self.dialog[turn_idx]['problem_data'] = p


class MultiLightModelChatOnboardWorld(ModelChatOnboardWorld):
    pass


def make_onboarding_world(opt, agent):
    return MultiLightModelChatOnboardWorld(opt, agent)


def make_world(opt, agents):

    # Extract important components from opt
    statistics_condition = opt['statistics_condition']
    context_generator = opt['context_generator']

    # Get context: personas, previous utterances, etc.
    if context_generator is not None:
        context_info = context_generator.get_context()
    else:
        context_info = None

    # Decide on a bot to use
    run_statistics = opt['run_statistics']
    with statistics_condition:
        remaining_counts_needed = [
            (m, c - run_statistics[m]) for (m, c) in opt['conversations_needed'].items()
        ]
        remaining_counts_needed.sort(reverse=True, key=lambda x: x[1])
        model_name = remaining_counts_needed[0][0]
        print(f'Remaining conversation counts needed: {remaining_counts_needed}')
        print(f'Choosing the "{model_name}" model for the bot.')
    bot_worker = get_bot_worker(opt=opt, model_name=model_name)

    return MultipartyChatWorld(
        opt, agent=agents[0], bot=bot_worker, context_info=context_info
    )


def get_world_params():
    return {"agent_count": 1}


def get_settings(opt=None):
    """
    Returns the conversation settings.

    This is a place-holder function with a few hand selected settings, override with
    more for real data collection.
    """
    return [
        {
            'personas': [
                {
                    'name': 'grass snake',
                    'persona': "I'm a grass snake. I slither around the castle and fields. I eat the rodents that eat the grain.",
                },
                {
                    'name': 'tribesman',
                    'persona': (
                        "I am a tribesman in my group. I am known as a leader in my community and love to help my people. "
                        " I'm very level headed and don't get angry easily. Many of my peers come to me to solve disagreements."
                    ),
                },
                {
                    'name': 'thief',
                    'persona': (
                        'I live alone in a tent in the woods. I steal food from the townspeople and coal from the blacksmith.'
                        ' The village police can not find me to put me in jail.'
                    ),
                },
            ],
            'location': {
                'name': 'Bamboo hut',
                'description': (
                    "Built of bamboo trunks and a bamboo leaf roof, this small hut has one window on each side and a short door,"
                    " where those who enter must stoop down so they don't hit their heads. "
                    "A dirt floor is covered with palm fronds gathered from the jungle; "
                    "four small rocks are placed around the center of the room, forming a place for the occupants to sit. "
                    "A small fire burns just outside of the hut, and a wooden spit is suspended over the fire. "
                    "One of the support poles of the hut has a woven grass bag hanging from it. The bag contains a half-dozen coconuts,"
                    " clearly gathered for consuming at a later time. A colorful lizard is sleeping in the sun in one of the windows."
                ),
            },
        },
        {
            'personas': [
                {
                    'name': 'clergy',
                    'persona': (
                        "I oversee the castle's chapel.  I collect alms for the poor. "
                        "I am the spiritual leader of the subjects of the kingdom."
                    ),
                },
                {
                    'name': 'Nuns',
                    'persona': (
                        "I am a nun and I live in a monastery with others nuns and fathers who server the king."
                        " I pray to the lord everyday that Queen remains in good health. "
                        "I was made a sister at a young age and didn't have a choice. "
                        "I will never know what being with a man will feel like."
                    ),
                },
                {
                    'name': 'priest',
                    'persona': 'I am here to help the needy. I am well respected in the town. I can not accept lying.',
                },
            ],
            'location': {
                'name': 'Church Entryway',
                'description': (
                    'The church has marble floors and a huge frost window. '
                    'There are benches made from wood and a big organ can be seen at the front stage.'
                    ' There is gold trim all around the church.'
                ),
            },
        },
    ]


class ContextGenerator:
    """
    Generates contexts shown to crowdsourced workers during data collection.
    """

    def __init__(self, opt, datatype: str = 'test', seed: int = None):
        """
        Initalize the context generator.
        """
        if seed is not None:
            self.rng = random.Random(seed)
        else:
            self.rng = random.Random()

    def get_context(self) -> dict:
        """
        Get context information to be shown at the beginning of one conversation. Values
        in return dict:

        - context_dataset: the dataset
        - personas: a list of dict where each dictionary is a persona as stored in this task messages.
        """
        setting = random.choice(get_settings())
        return {
            'context_dataset': 'multi-modelchat',
            'personas': setting['personas'],
            'location': setting['location'],
        }



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Base script for running TOD model-model chats.

For example, to extract gold ground truth data from the holdout version of Google SGD, run

```
python -u -m parlai.scripts.tod_world_script --api-schema-grounding-model parlai.tasks.google_sgd_simulation_splits.agents:OutDomainApiSchemaAgent --goal-grounding-model parlai.tasks.google_sgd_simulation_splits.agents:OutDomainGoalAgent --user-model parlai.tasks.google_sgd_simulation_splits.agents:OutDomainUserUttAgent --system-model parlai.tasks.google_sgd_simulation_splits.agents:OutDomainApiCallAndSysUttAgent --api-resp-model parlai.tasks.google_sgd_simulation_splits.agents:OutDomainApiResponseAgent -dt valid --num-episodes -1 --episodes-randomization-seed 42 --world-logs gold-valid
```

This file handles
1. Script param setup, including that used for loading agents which may have their own parameters
2. Running the world (including handling batching, until num episodes or length of epoch has been met).
3. File I/O for both reports (for metrics) and conversation logs + logic for displaying prints
"""

import json
from copy import deepcopy
from shutil import copyfile
import os

import parlai.utils.logging as logging
import parlai.core.tod.tod_world as tod_world
import parlai.core.tod.tod_agents as tod_world_agents
from parlai.core.agents import create_agent
from parlai.core.metrics import dict_report, aggregate_unnamed_reports
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript, register_script
from parlai.utils.distributed import (
    is_primary_worker,
    all_gather_list,
    is_distributed,
    get_rank,
    sync_object,
    num_workers,
)
from parlai.utils.io import PathManager
from parlai.utils.misc import TimeLogger, nice_report
from parlai.utils.world_logging import WorldLogger


class TodWorldLogger(WorldLogger):
    """
    WorldLogger has most of what we need.

    We could if-class this logic in it directly, but inheritence + override here is
    neater.
    """

    def _is_batch_world(self, world):
        return True

    def _log_batch(self, world):
        batch_acts = world.get_batch_acts()
        for i, acts in enumerate(batch_acts):
            acts = [
                act for act in acts if act is not None and "id" in act and "text" in act
            ]
            acts = [
                act
                for act in acts
                if act["id"] != "" and (act["text"] != "" or "Human" in act["id"])
            ]
            if len(acts) > 0:
                self._add_msgs(acts, idx=i)
            if world.episode_done():
                self.reset_world(idx=i)


class TodWorldParser(ParlaiParser):
    def add_extra_args(self, args=None):
        super().add_extra_args(args)
        parsed = vars(self.parse_known_args(args, nohelp=True)[0])
        # Also load extra args options if a file is given.
        if parsed.get("init_opt") is not None:
            try:
                self._load_known_opts(parsed.get("init_opt"), parsed)
            except FileNotFoundError:
                # don't die if -o isn't found here. See comment in second call
                # later on.
                pass
        parsed = self._infer_datapath(parsed)

        partial = Opt(parsed)

        for model in [
            "system_model",
            "user_model",
            "api_schema_grounding_model",
            "goal_grounding_model",
            "api_resp_model",
        ]:
            if (
                model in partial
                and partial[model] is not None
                and len(partial[model]) > 0
            ):
                self.add_model_subargs(partial[model], partial)

        for model_file_prefix in ["system", "user"]:
            key = model_file_prefix + "_model_file"
            if key in partial and partial[key] and len(partial[key]) > 0:
                model_name = self._get_model_name_from_model_file(key, partial)
                self.add_model_subargs(model_name, partial)

    def _get_model_name_from_model_file(self, key, opt):
        """
        Get the model name from either `--model` or `--model-file`.
        """
        # try to get model name from model opt file
        model_file = opt.get(key, None)
        optfile = model_file + ".opt"
        new_opt = Opt.load(optfile)
        model = new_opt.get("model", None)
        return model


@register_script("tod_world_script")
class TodWorldScript(ParlaiScript):
    @classmethod
    def setup_tod_args(cls, parser: ParlaiParser):
        tod_args = parser.add_argument_group(
            "TOD World Script Agent arguments. NOTE: If there are issues with invoking downstream opts of agents specified here sometimes you will have more luck with `python -u -m parlai.scripts.tod_world_script` than `parlai tod_world_script`."
        )
        tod_args.add_argument(
            "--system-model-file",
            default="",
            help="Define the system model for the chat. Exactly one of this or system-model must be specified",
        )

        tod_args.add_argument(
            "--system-model",
            default="",
            help="Define the system agent for the chat. Exactly one of this or system-model-file must be specified",
        )

        tod_args.add_argument(
            "--user-model-file",
            default="",
            help="Define the user model for the chat. Exactly one of this user-model must be specified. Currently assumed to be the API Call creation agent as well.",
        )

        tod_args.add_argument(
            "--user-model",
            default="",
            help="Define the user agent for the chat. Exactly one of this or user-model-file must be specified. Currently assumed to be the API Call creation agent as well.",
        )

        tod_args.add_argument(
            "--api-resp-model",
            default="",
            help="Agent used for defining API response values",
        )

        tod_args.add_argument(
            "--api-schema-grounding-model",
            default="",
            help="Agent used in first turn to grounding api call/response agents with api schemas. Will use EmptyApiSchemaAgent if both this and `--api-schemas` not set.",
        )

        tod_args.add_argument(
            "--goal-grounding-model",
            default="",
            help="Agent used in first turn to grounding user agent with goal. Will use EmptyGoalAgent if not set",
        )

        tod_args.add_argument(
            "--api-schemas",
            default=None,
            help="If set and `--api-schema-grounding-model` is empty, will infer `--api-schema-grounding-model` based on this and a regex on `--goal-grounding-model`. If you run into issues with parsing order of opts using this flag, just switch to `--api-schema-grounding-model`.",
        )

    @classmethod
    def setup_args(cls):
        # Use default parlai args for logging + the like, but don't need model args since we specify those manually via command-line
        parser = TodWorldParser(
            True, False, "World for chatting with the TOD conversation structure"
        )
        # Following params are same as the `eval_model` script
        parser.add_argument(
            "--report-filename",
            type=str,
            help="Saves a json file of the evaluation report either as an "
            'extension to the model-file (if begins with a ".") or a whole '
            "file path. Set to the empty string to not save at all.",
        )
        parser.add_argument(
            "--world-logs",
            type=str,
            help="Saves a jsonl file containing all of the task examples and "
            "model replies.",
        )
        parser.add_argument(
            "--save-format",
            type=str,
            default="conversations",
            choices=["conversations", "parlai"],
        )
        parser.add_argument(
            "--num-episodes",
            type=int,
            default=10,
            help="Number of episodes to display. Set to -1 for infinity or the number of examples of the first agent with a non-unlimited number of episodes in the world.",
        )
        parser.add_argument("-d", "--display-examples", type="bool", default=False)
        parser.add_argument("-ltim", "--log-every-n-secs", type=float, default=10)
        TodWorldLogger.add_cmdline_args(parser)

        # Following are specific to TOD World
        parser.add_argument(
            "--max-turns",
            type=int,
            default=30,
            help="The max number of full turns before chat ends, excluding prompting",
        )
        TodWorldScript.setup_tod_args(parser)

        return parser

    def _get_file_or_model_specifiable_agent(self, prefix, opt):
        if len(opt.get(f"{prefix}_model_file", "")) > 0:
            if len(opt.get(f"{prefix}_model", "")) > 0:
                raise KeyError(
                    "Both `--{prefix}-model-file` and `--{prefix}-model` specified. Exactly one should be."
                )
            model = self._make_agent(
                opt,
                f"{prefix}_model_file",
                requireModelExists=True,
                opt_key="model_file",
            )
        elif len(opt.get(f"{prefix}_model", "")) > 0:
            model = self._make_agent(opt, f"{prefix}_model", "")
        else:
            raise KeyError(
                f"Both `--{prefix}-model-file` and `--{prefix}-model` specified. Neither currently set."
            )
        return model

    def _get_model_or_default_agent(self, opt, key, default_class):
        if len(opt.get(key, "")) > 0:
            return self._make_agent(opt, key)
        return default_class(opt)

    def _get_tod_agents(self, opt: Opt):
        agents = [None] * tod_world.AGENT_COUNT

        agents[tod_world.USER_UTT_IDX] = self._get_file_or_model_specifiable_agent(
            "user", opt
        )

        # Get system agent, nothing that api call agent currently same as system agent
        system_model = self._get_file_or_model_specifiable_agent("system", opt)
        agents[tod_world.SYSTEM_UTT_IDX] = system_model
        agents[tod_world.API_CALL_IDX] = system_model

        agents[tod_world.API_RESP_IDX] = self._make_agent(opt, "api_resp_model")
        agents[tod_world.GOAL_GROUNDING_IDX] = self._get_model_or_default_agent(
            opt, "goal_grounding_model", tod_world_agents.EmptyGoalAgent
        )

        if "api_schema_grounding_model" not in opt and "api_schemas" in opt:
            opt["api_schema_grounding_model"] = opt.get(
                "goal_grounding_model", ""
            ).replace("Goal", "ApiSchema")

        agents[tod_world.API_SCHEMA_GROUNDING_IDX] = self._get_model_or_default_agent(
            opt, "api_schema_grounding_model", tod_world_agents.EmptyApiSchemaAgent
        )

        return agents

    def _make_agent(self, opt_raw, name, requireModelExists=False, opt_key="model"):
        """
        Hack.

        `create_agent` expects opt[`model`] to specify the model type and we're
        specifying multiple models from other opt arguments (ex.
        `system_model`/`user_model` etc), so this swaps it in.
        """
        opt = deepcopy(opt_raw)
        opt[opt_key] = opt[name]
        print(opt_key, name)
        return create_agent(opt, requireModelExists)

    def _run_episode(self, opt, world, world_logger):
        while not world.episode_done():
            world.parley()
            world_logger.log(world)

            if opt["display_examples"]:
                logging.info(world.display())

        if opt["display_examples"]:
            logging.info("-- end of episode --")

        world.reset()
        world_logger.reset_world()  # flush this episode
        return zip(world.get_last_batch_goals(), world.get_last_batch_episode_metrics())

    def _save_outputs(self, opt, world, logger, episode_metrics):
        if is_distributed():  # flatten everything intelligently if need be
            world_report = aggregate_unnamed_reports(all_gather_list(world.report()))
            episode_metrics_unflattened = all_gather_list(episode_metrics)
            flattened = []
            for rank_elem in episode_metrics_unflattened:
                for elem in rank_elem:
                    flattened.append(elem)
            episode_metrics = flattened
        else:
            world_report = world.report()
        logging.report("Final report:\n" + nice_report(world_report))

        report = dict_report(world_report)

        def get_episode_report(goal, episode_metric):
            metrics_dict = dict_report(episode_metric.report())
            metrics_dict["goal"] = goal
            return metrics_dict

        report["tod_metrics"] = [get_episode_report(g, e) for g, e in episode_metrics]

        if "report_filename" in opt and opt["report_filename"] is not None:
            if len(world_report) == 0:
                logging.warning("Report is empty; not saving report")

            report_fname = f"{opt['report_filename']}.json"
            # Save report
            if not is_distributed() or is_primary_worker():
                with PathManager.open(report_fname, "w") as f:
                    logging.info(f"Saving model report to {report_fname}")
                    json.dump({"opt": opt, "report": report}, f, indent=4)
                    f.write("\n")  # for jq

        if "world_logs" in opt and opt["world_logs"] is not None:
            if is_distributed():  # Save separately, then aggregate together
                rank = get_rank()
                log_outfile_part = (
                    f"{opt['world_logs']}_{opt['save_format']}_{rank}.jsonl"
                )
                logger.write(log_outfile_part, world, file_format=opt["save_format"])
                sync_object(None)
                if is_primary_worker():
                    log_outfile = f"{opt['world_logs']}_{opt['save_format']}.jsonl"
                    log_outfile_metadata = (
                        f"{opt['world_logs']}_{opt['save_format']}.metadata"
                    )
                    with open(log_outfile, "w+") as outfile:
                        for rank in range(num_workers()):
                            log_outfile_part = (
                                f"{opt['world_logs']}_{opt['save_format']}_{rank}.jsonl"
                            )
                            with open(log_outfile_part) as infile:
                                for line in infile:
                                    json_blob = json.loads(line.strip())
                                    if (
                                        len(json_blob["dialog"]) < 2
                                    ):  # skip when we don't have generation
                                        continue
                                    json_blob["metadata_path"] = log_outfile_metadata
                                    outfile.write(json.dumps(json_blob))
                                    outfile.write("\n")
                            log_output_part_metadata = f"{opt['world_logs']}_{opt['save_format']}_{rank}.metadata"
                            if rank == 0:
                                copyfile(
                                    log_output_part_metadata, log_outfile_metadata
                                ),
                            os.remove(log_outfile_part)
                            os.remove(log_output_part_metadata)
            else:
                log_outfile = f"{opt['world_logs']}_{opt['save_format']}.jsonl"
                logger.write(log_outfile, world, file_format=opt["save_format"])

        return report

    def _setup_world(self):
        # setup world, manually finaggling necessary opt info as needed
        self.opt["task"] = "TodWorld"
        world = tod_world.TodWorld(self.opt, agents=self._get_tod_agents(self.opt))
        return world

    def run(self):
        opt = self.opt

        world = self._setup_world()
        logger = TodWorldLogger(opt)

        # set up logging
        log_every_n_secs = opt.get("log_every_n_secs", -1)
        if log_every_n_secs <= 0:
            log_every_n_secs = float("inf")
        log_time = TimeLogger()

        # episode counter
        max_episodes = opt.get("num_episodes", -1)
        if max_episodes < 0:
            max_episodes = float("inf")
        world_num_episodes = world.num_episodes()
        if world_num_episodes > 0:
            max_episodes = min(max_episodes, world_num_episodes)

        ep_count = 0
        episode_metrics = []
        while not world.epoch_done() and ep_count < max_episodes:
            episode_metrics.extend(self._run_episode(opt, world, logger))
            ep_count += opt.get("batchsize", 1)
            if log_time.time() > log_every_n_secs:
                report = world.report()
                text, report = log_time.log(ep_count, max_episodes, report)
                logging.info(text)

        return self._save_outputs(opt, world, logger, episode_metrics)


if __name__ == "__main__":
    TodWorldScript.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from typing import Dict, Optional, Tuple

from parlai.tasks.reasoning.base import (
    t_REASON_PREFIX_TOKEN,
    t_REASON,
    AbstractReason,
)


class FreeFormReason(AbstractReason):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group("FreeFormReason args")
        group.add_argument(
            "--reason-token",
            default="REASON: ",
            type=str,
        )
        return parser

    def get_full_reason_text(
        self, example_dict
    ) -> Tuple[t_REASON_PREFIX_TOKEN, t_REASON, Dict]:
        return (
            self.opt["reason_token"],
            example_dict["reason"],
            {"reason", example_dict["reason"]},
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import random
from typing import Optional
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import FixedDialogTeacher
from parlai.tasks.dialogue_safety.agents import StandardTeacher, NOT_OK_CLASS, OK_CLASS
from parlai.utils.io import PathManager
from .build import build, FILE_TYPE_EXTENSIONS, TROLL_TYPES
import os


def _path(opt):
    # build the data if it does not exist
    build(opt['datapath'])
    # set up path to data (specific to each dataset)
    data_path = os.path.join(opt['datapath'], 'safety_mix')
    return data_path


class SafetyMixTeacher(FixedDialogTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = StandardTeacher.add_cmdline_args(parser, partial_opt=partial_opt)
        parser = parser.add_argument_group('SafetyMix arguments')
        parser.add_argument(
            '--mix-user-type',
            type=str,
            default='troll',
            help=f'The troll user type you want in the safety mix. Possible options are: {TROLL_TYPES}',
        )
        return parser

    def __init__(self, opt, shared=None):
        self.opt = opt
        dpath = _path(opt)
        self.data_path = dpath
        self.troll_type = opt['mix_user_type']
        assert self.troll_type in TROLL_TYPES, f'{self.troll_type} not supported.'

        self.fixed_random = random.Random(42)
        self.label_candidates = [NOT_OK_CLASS, OK_CLASS]

        if shared and 'data' in shared:
            self.data = shared['data']
        else:
            self._setup_data(opt['datatype'])
        super().__init__(opt, shared)
        self.reset()

    def _load_data_dump(self, datatype):
        d_type = datatype.split(':')[0]
        loaded_data = []
        with PathManager.open(
            os.path.join(
                self.data_path, f'{self.troll_type}{FILE_TYPE_EXTENSIONS[d_type]}'
            ),
            'rb',
        ) as f:
            dump = list(f)
        for json_str in dump:
            loaded_data.append(json.loads(json_str))
        return loaded_data

    def _setup_data(self, datatype):
        # load data
        self.data_dump = self._load_data_dump(datatype)
        self.data = self.data_dump
        self.fixed_random.shuffle(self.data)

    def num_episodes(self):
        return len(self.data)

    def num_examples(self):
        return len(self.data)

    def get(self, episode_idx, entry_idx):
        return Message(self.data[episode_idx])

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared


class PosSafetyMixTeacher(SafetyMixTeacher):
    def _load_data_dump(self, datatype):
        d_type = datatype.split(':')[0]
        loaded_data = []
        with PathManager.open(
            os.path.join(
                self.data_path, f'pos_{self.troll_type}{FILE_TYPE_EXTENSIONS[d_type]}'
            ),
            'rb',
        ) as f:
            dump = list(f)
        for json_str in dump:
            loaded_data.append(json.loads(json_str))
        return loaded_data


class NegSafetyMixTeacher(SafetyMixTeacher):
    def _load_data_dump(self, datatype):
        d_type = datatype.split(':')[0]
        loaded_data = []
        with PathManager.open(
            os.path.join(
                self.data_path, f'neg_{self.troll_type}{FILE_TYPE_EXTENSIONS[d_type]}'
            ),
            'rb',
        ) as f:
            dump = list(f)
        for json_str in dump:
            loaded_data.append(json.loads(json_str))
        return loaded_data


class DefaultTeacher(SafetyMixTeacher):
    pass
